
/* Portable Object Compiler */
/* $Id: objc1.m,v 1.46 1998/10/25 12:21:25 stes Exp $ */

/*
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "objc1.h"
#include "oclib.h"
#include <string.h>		/* strtok */
#include <stdlib.h>		/* getenv */

/*
 * Most important stuff first.
 */

static void 
version(void)
{
  printf("1.10.2\n");
  exit(0);
}

static void 
copyright(void)
{
  STR s1, s2;

  s1 = "Portable Object Compiler 1.10.2 (c) 1997, 98.\n";
  s2 = "Distributed under the terms of the GNU LGPL.\n";
  printf(s1);
  printf(s2);
}

/*
 * Functions to parse command line arguments.
 */

static void 
unknownOption(id option)
{
  STR name = ([options cplus]) ? "objcpls1" : "objc1";

  fprintf(stderr, "%s: Illegal option %s\n", name, [option str]);
  exit(1);
}

static void 
illegalArgument(id option, id arg)
{
  char *msg = "%s: Illegal argument %s for %s\n";
  STR name = ([options cplus]) ? "objcpls1" : "objc1";

  fprintf(stderr, msg, name, [arg str], [option str]);
  exit(1);
}

/*
 * Use 'dash' character to distinguish between options and filenames.
 */

static BOOL 
isOption(id s)
{
  return [s size] && [s charAt:0] == '-';
}

/* reading options from string or file */

static id 
optionsFromString(id aCltn, char *s)
{
  char *p;
  char *delims = " \t\n\r";
  id buffer = [String str:s];

  p = strtok([buffer str], delims);

  while (p != NULL) {
    [aCltn add:[String str:p]];
    p = strtok(NULL, delims);
  }

  return aCltn;
}

static id 
optionsFromFile(id aCltn, FILE * f)
{
  char buf[BUFSIZ + 1];

  while (!feof(f)) {
    if (fgets(buf, BUFSIZ, f)) {
      /* this works 'cause strtok also deletes \n */
      optionsFromString(aCltn, buf);
    }
  }

  return aCltn;
}

/* this is mostly for MS-DOS cmd line length limitations */

static id 
optionsNamed(id aCltn, STR s)
{
  STR t;
  FILE *f;

  /* maybe there's such an environment variable */
  if ((t = getenv(s)))
    return optionsFromString(aCltn, t);

  /* maybe there's such a file */
  if ((f = fopen(s, "r"))) {
    aCltn = optionsFromFile(aCltn, f);
    fclose(f);
    return aCltn;
  }
  /* nope */
  return nil;
}

/*
 * Cmd Line Options
 */

static id 
cmdLineOptions(int argc, char *argv[])
{
  int i;
  id aCltn = [OrdCltn new];

  for (i = 1; i < argc; i++)
    [aCltn add:[String str:argv[i]]];
  return aCltn;
}

/*
 * Filename for first line tag output (source filename)
 * (which we want to be file.m instead of file.P)
 */

static int 
setfilename(id option, id args)
{
  id arg = [args next];

  if (isOption(arg))
    illegalArgument(option, arg);
  [options srcfilename:arg];
  return 0;
}

/*
 * Make initialization function customazible.
 */

static int 
initCall(id option, id args)
{
  id arg = [args next];

  if (isOption(arg))
    illegalArgument(option, arg);
  [options initCall:arg];
  return 0;
}

static int 
mainfun(id option, id args)
{
  id arg = [args next];

  if (isOption(arg))
    illegalArgument(option, arg);
  [options mainfun:arg];
  return 0;
}

static int 
bind(id option, id args)
{
  id arg = [args next];

  if (isOption(arg))
    illegalArgument(option, arg);
  [options bind:arg];
  return 0;
}

/*
 * Make line max settable (works only for identifiers now).
 */

static int 
linemax(id option, id args)
{
  id arg = [args next];

  if (isOption(arg))
    illegalArgument(option, arg);
  [options linemax:[arg asInt]];
  return 0;
}

/*
 * @ filename is understood to take options from "filename"
 * Maybe we should also allow @filename (without space)
 */

static void setOptions(id aCltn);

static int 
setOptionsFromFile(id option, id args)
{
  id arg = [args next];

  if (isOption(arg))
    illegalArgument(option, arg);
  setOptions(optionsNamed([OrdCltn new], [arg str]));
  return 0;
}

/*
 * Builtin functions (a la __builtin_isfloat).
 */

static int 
addBuiltinType(id option, id args)
{
  id arg = [args next];

  if (isOption(arg))
    illegalArgument(option, arg);
  [[[Identifier str:[arg str]] markdecl:YES] deftypename];
  return 0;
}

static int 
addBuiltinFun(id option, id args)
{
  id arg = [args next];

  if (isOption(arg))
    illegalArgument(option, arg);
  [[Identifier str:[arg str]] defbuiltinfun];
  return 0;
}

/*
 * ANSI C book says #line is the directive.
 * GNU cc also accepts # lineno
 */

static void 
setShortTagFormat(void)
{
  id arg = [String str:"# %d \"%s\""];

  [options tagformat:arg];
}

/*
 * Function to parse command line arguments.
 */

static void 
setOptions(id aCltn)
{
  id args, s;
  int fileCount = 0;		/* optional inFile and outFile */
  BOOL checkOption = YES;	/* YES if filename can't begin with a dash */

  args = [aCltn eachElement];

  while ((s = [args next])) {
    if ([s isEqualSTR:"-quiet"]) {
      [options quiet:YES];
    } else if ([s isEqualSTR:"-z"]) {
      [options outputCode:NO];
    } else if ([s isEqualSTR:"@"]) {
      setOptionsFromFile(options, args);
    } else if ([s isEqualSTR:"-u"]) {
      [options buffered:NO];
    } else if ([s isEqualSTR:"-version"]) {
      [options version:YES];
    } else if ([s isEqualSTR:"-otb"]) {
      [options otb:YES];
    } else if ([s isEqualSTR:"-gnu"]) {
      [options gnu:YES];
    } else if ([s isEqualSTR:"-noasm"]) {
      [options asmKeyword:NO];
    } else if ([s isEqualSTR:"-nolonglong"]) {
      [options llkeyword:NO];
    } else if ([s isEqualSTR:"-msdos"]) {
      [options msdos:YES];
    } else if ([s isEqualSTR:"-watcom"]) {
      [options watcom:YES];
    } else if ([s isEqualSTR:"-comments"]) {
      [options comments:YES];
    } else if ([s isEqualSTR:"-ibmvac"]) {
      [options ibmvac:YES];
    } else if ([s isEqualSTR:"-noSelTbl"]) {
      [options seltranslation:NO];
    } else if ([s isEqualSTR:"-noCategories"]) {
      [options categories:NO];
    } else if ([s isEqualSTR:"-noBlocks"]) {
      [options blocks:NO];
    } else if ([s isEqualSTR:"-fwd"]) {
      [options fwd:YES];
    } else if ([s isEqualSTR:"-noSelPtr"]) {
      [options selptr:NO];
    } else if ([s isEqualSTR:"-noFwd"]) {
      [options fwd:NO];
    } else if ([s isEqualSTR:"-noCache"]) {
      [options cache:NO];
    } else if ([s isEqualSTR:"-noFiler"]) {
      [options filerCode:NO];
    } else if ([s isEqualSTR:"-noSelfAssign"]) {
      [options selfAssign:NO];
    } else if ([s isEqualSTR:"-noNilRcvr"]) {
      [options nilRcvr:NO];
    } else if ([s isEqualSTR:"-objc"]) {
      [options genCode:NO];
    } else if ([s isEqualSTR:"-cplus"]) {
      [options setCplus:YES];
    } else if ([s isEqualSTR:"-inlinecache"]) {
      [options inlineCache:YES];
    } else if ([s isEqualSTR:"-refBind"]) {
      [options refBind:YES];
    } else if ([s isEqualSTR:"-export"]) {
      bind(s, args);
    } else if ([s isEqualSTR:"-dllexport"]) {
      [options bind:[String str:"__declspec(dllexport)"]];
    } else if ([s isEqualSTR:"-traditional"]) {		/* ignore */
    } else if ([s isEqualSTR:"-filename"]) {
      setfilename(s, args);
    } else if ([s isEqualSTR:"-longlinetag"]) {		/* compatib. */
    } else if ([s isEqualSTR:"-shortTags"]) {
      setShortTagFormat();
    } else if ([s isEqualSTR:"-init"]) {
      initCall(s, args);
    } else if ([s isEqualSTR:"-noShared"]) {
      [options shareddata:NO];
    } else if ([s isEqualSTR:"-main"]) {
      mainfun(s, args);
    } else if ([s isEqualSTR:"-linemax"]) {
      linemax(s, args);
    } else if ([s isEqualSTR:"-builtinfunction"]) {
      addBuiltinFun(s, args);
    } else if ([s isEqualSTR:"-builtintype"]) {
      addBuiltinType(s, args);
    } else if ([s isEqualSTR:"-nolinetags"]) {
      [options linetags:NO];
    } else if ([s isEqualSTR:"-oneperfile"]) {
      [options onePfile:YES];
    } else if ([s isEqualSTR:"-wClassUsedAsType"]) {
      [options warnclasstype:NO];
    } else if ([s isEqualSTR:"-wTypeConflict"]) {
      [options warntypeconflict:NO];
    } else if ([s isEqualSTR:"-wLocalInstance"]) {
      [options warnlocalnst:NO];
    } else if ([s isEqualSTR:"-wUndefinedMethod"]) {
      [options warnundefined:NO];
    } else if ([s isEqualSTR:"-wInterfaceNotFound"]) {
      [options warnnotfound:NO];
    } else if ([s isEqualSTR:"-WLex"]) {
      [options warnlex:YES];
    } else if ([s isEqualSTR:"-postlink"]) {
      [options postlink:YES];
    } else if ([s isEqualSTR:"-checkbindfunction"]) {
      [options checkBind:YES];
    } else if ([s isEqualSTR:"-debugInfo"]) {
      [options debugInfo:YES];
    } else if ([s isEqualSTR:"-ppi"]) {
      [options ppi:YES];
    } else if ([s isEqualSTR:"-"]) {
      checkOption = NO;
    } else if (checkOption && isOption(s)) {
      unknownOption(s);
    } else if (fileCount == 0) {
      [options inFile:s];
      fileCount++;
    } else if (fileCount == 1) {
      [options setOutFile:s];
      fileCount++;
    } else {
      unknownOption(s);
    }
  }
}

/*
 * Opening and closing input file.  We're not dealing with wrapping around
 * files (include)  (because we're processing cpp output)
 */

extern FILE *yyin;		/* scan.l global */

static void 
openInFile(id inFile)
{
  if (inFile) {
    yyin = openFile([inFile str], "r");
  } else {
    /* stdin */
  }

  /* initialize LineTag mechanism */
  /* everytime a newline is encountered, the scanner */
  /* will up the current LineTag object */

  [LineTag new];
}

static void 
closeInFile(id inFile)
{
  if (inFile) {
    fclose(yyin);
  } else {
    /* stdin */
  }
}

/*
 * Output code.
 */

static id outStream;

static void 
openOutFile()
{
  id outFile = [options outFile];

  if (outFile) {
    outStream = [OutputStream str:[outFile str]];
  } else {
    outStream = [OutputStream new];	/* stdout */
  }

  /* output a first LineTag */
  /* initialize output mechanism for LineTags */
  [LineTag outputOn:outStream];
}

static void 
closeOutFile()
{
  [LineTag finishOn:outStream];
  [outStream close];
}

/*
 * After each declaration (top level definition), processNode()
 * is called to do semantical analysis (defining types etc.)
 * and then outputNode() to generate some code (if anything).
 */

void 
processNode(id aNode)
{
  if (aNode)
    [aNode process];
}

void 
outputNode(id aNode)
{
  BOOL doOutput = [options outputCode];
  BOOL doTranslate = [options genCode];

  if (doOutput && aNode != nil) {
    if (doTranslate) {

      /* if node contains any blocks, output block definitions first */
      if ([options blocks]) {
	[aNode processBlocks];
	[aNode genblockdefsOn:outStream];
      }
      /* do code translation objc -> c */

      [aNode genOn:outStream];
    } else {
      /* just dump objc to outstream */
      [aNode outputOn:outStream];
    }
  }
}


/*
 * Main entrance point of program.
 */

extern int yyparse();		/* parse.y entry point */

void 
main(int argc, char *argv[])
{
  [Options new];		/* options is a global */

  setOptions(cmdLineOptions(argc, argv));

  if ([options version]) {
    version();
  }
  if (![options quiet]) {
    copyright();
  }
  openOutFile();
  openInFile([options inFile]);

  [Module prologueOn:outStream];
  if (yyparse())
    exit(1);
  [Module epilogueOn:outStream];

  closeInFile([options inFile]);
  closeOutFile();

  exit(exitStatus);
}



