
/*
 * Portable Object Compiler (c) 1997.  All Rights Reserved.
 * $Id: classdef.m,v 1.20 1998/10/06 15:23:43 stes Exp $
 */

/*
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include "classdef.h"
#include "options.h"
#include "output.h"
#include "linetag.h"
#include "module.h"
#include "identif.h"
#include "set.h"
#include "dictnary.h"
#include "ordcltn.h"
#include "ocstring.h"
#include "methinfo.h"
#include "methimpl.h"
#include "methdecl.h"
#include "msgxpr.h"
#include "compstat.h"
#include "token.h"
#include <assert.h>

/*  class implementation (or interface) currently being parsed */
static id curimpl;

static id curitypes;
static id curivars;
static id curctypes;
static id curcvars;
static id curivarnodes;
static id curcvarnodes;

static id grepTypes;		/* for parsing */
static id grepVars;

static id currentInterface;	/* for method translation */

static id allInterfaces;

@implementation ClassDefinition
- free
{
  return nil;
}


+ findInterface:aName
{
  if (!allInterfaces)
    allInterfaces = [Dictionary new];
  return [allInterfaces atKey:aName];
}

- addInterface
{
  assert(allInterfaces);
  [allInterfaces atKey:name put:self];
  return self;
}


- idName
{
  return name;
}

static void 
warnMultiple(id aName)
{
  char b[8192];
  char *s = [aName str];

  sprintf(b, "Found multiple interfaces for \"%s\"", s);
  warnError(b);
}

static void 
warnnotfound(id aName)
{
  char b[8192];
  char *s = [aName str];

  sprintf(b, "Cannot find interface declaration for \"%s\"", s);
  warnError(b);
}

- (BOOL) isimpltoken
{
  return [[subnodes at:0] isEqualSTR:"@implementation"];
}

- (BOOL) isintftoken
{
  return [[subnodes at:0] isEqualSTR:"@interface"];
}

- processName:myName superName:hisName
{
  BOOL intfToken = [self isintftoken];
  BOOL implToken = [self isimpltoken];

  /* get string object from identifier */
  /* myName can be CLASSNAME_TOKEN or simply IDENTIFIER_TOKEN */
  /* (depending on whether it was defined before or not) */

  name = [myName string];

  /* interface is nil if class not yet defined */
  interface = [ClassDefinition findInterface:name];

  if (interface) {
    if (intfToken) {
      warnMultiple(myName);
    } else {
      isImplementation = YES;
      interfaceEmitted = YES;
    }
  } else {
    /* this is the case of either an interface declaration
     * or the case of an implementation that is at the same
     * time its own interface
     */

    if (!intfToken)
      [self isImplementation:YES];

    /* issue a warning, but not for '=' class defs */
    if (implToken && [options warnnotfound]) {
      warnnotfound(myName);
    }
    /* make the identifier a special token (CLASSNAME_TOKEN) */
    [myName defclassname];

    /* declare class interface, set interface to self */
    interface = [self addInterface];
  }

  if (myName != nil && hisName != nil && [hisName isEqual:myName]) {
    fatalError("Name of class and superclass equal");
    return self;
  }
  /* get string object from identifier */
  superName = (hisName) ? [hisName string] : nil;

  if (interface != self) {
    /* interface superName can be nil for rootclass */
    if ([interface superName]) {
      if (superName) {
	if (![superName isEqual:[interface superName]])
	  fatalError("Found conflicting superclass");
      }
    } else {
      if (superName)
	fatalError("Found conflicting superclass");
    }
  }
  return self;
}

- superName
{
  return superName;
}

- (BOOL) isCategory
{
  return isCategory;
}

- isCategory:(BOOL)flag
{
  isCategory = flag;
  return self;
}

/* Very primitive mechanism to collect the instance variables.
 * We need them to expand something like "ivar" to "self->ivar",
 * but at this moment, just remember the name, and a node
 * representation of the type.
 */

+ declareIvar:aNode Itype:bNode
{
  if (!grepVars)
    grepVars = [OrdCltn new];
  if (!grepTypes)
    grepTypes = [OrdCltn new];
  [aNode grepIvars:grepVars addItype:bNode to:grepTypes];
  return self;
}

- totalIvars
{
  if (totalIvars) {
    return totalIvars;
  } else {
    id face;

    /* not an OrdCltn  - idea is to have fast lookup (contains:) */
    totalIvars = [Set new];
    for (face = self; face; face = [face superInterface]) {
      [totalIvars addContentsOf:[face allIvars]];
    }
    return totalIvars;
  }
}

- totalCvars
{
  if (totalCvars) {
    return totalCvars;
  } else {
    id face;

    /* not an OrdCltn - idea is to have fast lookup (contains:) */
    totalCvars = [Set new];
    for (face = self; face; face = [face superInterface]) {
      [totalCvars addContentsOf:[face allCvars]];
    }
    return totalCvars;
  }
}

+ (BOOL) inmethimpl
{
  return (currentInterface != nil);
}

+ (BOOL) iscurivar:anIdentifier
{
  if (currentInterface) {
    return [[currentInterface totalIvars] contains:anIdentifier];
  } else {
    /* outside method implementation, can still be in class impl */
    return NO;
  }
}

+ (BOOL) iscurcvar:anIdentifier
{
  if (currentInterface) {
    return [[currentInterface totalCvars] contains:anIdentifier];
  } else {
    /* outside method implementation, can still be in class impl */
    return NO;
  }
}

- processVars
{
  if (curivars) {
    ivarNodes = curivarnodes;
    allIvars = curivars;
    allItypes = curitypes;
    assert([allIvars size] == [allItypes size]);
    curivarnodes = nil;
    curivars = nil;
    curitypes = nil;

  } else {
    /* no new ivars in this subclass */
    ivarNodes = nil;
    allIvars = [OrdCltn new];
    allItypes = [OrdCltn new];
  }

  if (curcvars) {
    cvarNodes = curcvarnodes;
    allCvars = curcvars;
    allCtypes = curctypes;
    assert([allCvars size] == [allCtypes size]);
    curcvarnodes = nil;
    curcvars = nil;
    curctypes = nil;

  } else {
    /* no new cvars in this subclass */
    cvarNodes = nil;
    allCvars = [OrdCltn new];
    allCtypes = [OrdCltn new];
  }

  /* if duplicated in both @interface and @implementation */
  /* they have to literally match */

  if (interface != self && (curcvars || curivars)) {
    if (![ivarNodes isEqual:[interface ivarNodes]]) {
      fatalError("conflicting instance variable declaration");
    }
    if (![cvarNodes isEqual:[interface cvarNodes]]) {
      fatalError("conflicting class variable declaration");
    }
  }
  return self;
}

- superInterface
{
  if (superName) {
    id superInterface = [ClassDefinition findInterface:superName];

    assert(superInterface);
    return superInterface;
  } else {
    return nil;
  }
}

- rootInterface
{
  id rootface;
  id superface;

  rootface = self;
  while ((superface = [rootface superInterface]))
    rootface = superface;
  return rootface;
}

- rootName
{
  id rootface = [self rootInterface];

  assert(rootface);
  return [rootface idName];
}

+ grepIvarnodes:aNode
{
  curivarnodes = aNode;		/* for literal dumping as in @defs */
  curivars = grepVars;
  curitypes = grepTypes;
  grepVars = nil;
  grepTypes = nil;
  return self;
}

+ grepCvarnodes:aNode
{
  curcvarnodes = aNode;		/* for literal dumping as in @defs */
  curcvars = grepVars;
  curctypes = grepTypes;
  grepVars = nil;
  grepTypes = nil;
  return self;
}

- makeCurrent
{
  currentInterface = self;
  return self;
}

- allIvars
{
  return allIvars;
}

- allCvars
{
  return allCvars;
}

- allIdivars
{
  if (allIdivars) {
    return allIdivars;
  } else {
    int i, n;

    allIdivars = [OrdCltn new];
    for (i = 0, n = [allIvars size]; i < n; i++) {
      if ([[allItypes at:i] isId]) {
	[allIdivars add:[allIvars at:i]];
      }
    }
    return allIdivars;
  }
}

- ivarNodes
{
  return ivarNodes;
}

- cvarNodes
{
  return cvarNodes;
}

- clearCurrent
{
  currentInterface = nil;
  return self;
}

- ivarscommentOn:stream
{
  id ivar, ivars = [allIvars eachElement];

  [stream echo:"\n/*\n"];
  [stream echo:" * %s instance variables\n", [name str]];
  while ((ivar = [ivars next])) {
    [stream echo:" *"];
    [ivar taglessOn:stream];
    [stream newline];
  }
  [ivars free];
  [stream echo:" */\n"];
  return self;
}


- ivarnodesOn:stream
{
  if (superName)
    [[self superInterface] ivarnodesOn:stream];
  if (ivarNodes)
    [ivarNodes outputOn:stream];

  return self;
}

- cvarnodesOn:stream
{
  if (superName)
    [[self superInterface] cvarnodesOn:stream];
  if (cvarNodes)
    [cvarNodes outputOn:stream];

  return self;
}

- (STR) globalfun
{
  return [[Identifier str:[name str]] globalfun];
}

- (STR) structfun
{
  return [[Identifier str:[name str]] structfun];
}

- (STR) superfun
{
  return [[Identifier str:[name str]] superfun];
}

- (STR) metastructfun
{
  return [[Identifier str:[name str]] metastructfun];
}

- (STR) metasuperfun
{
  return [[Identifier str:[name str]] metasuperfun];
}

- externclassOn:stream
{
  STR e = [[options bind] str];	/* stuff like __export etc. */

  [stream newline];
  if ([options sharedData]) {
    [stream echo:"extern id %s %s;\n", e, [name str]];
  } else {
    /* yes, returns a pointer to the global, (id *) */
    [stream echo:"extern id %s *%s(void);\n", e, [self globalfun]];
  }
  return self;
}

- interfaceOn:stream
{
  [self externclassOn:stream];

  /* definition of our struct _SHARED */
  /* stes sep/2/97 this used to be in Module's prologue */
  /* moved to this location for class variables */

  /* in fact we only need struct shared of the meta, the class
   * being implemented and of its superclass, but just emitting
   * it for every interface is easier (and avoids duplication)
   */

  [self shartypeOn:stream];
  [stream newline];

  [LineTag forceTag];
  interfaceEmitted = YES;
  return self;
}


- isImplementation:(BOOL)flag
{
  isImplementation = flag;
  return self;
}

+ checkpending
{
  if (curimpl) {
    id aMsg = [curimpl idName];

    aMsg = [String sprintf:"Class %s not properly ended.", [aMsg str]];
    fatalError([aMsg str]);
    return nil;
  } else {
    return self;
  }
}

+ new
{
  [self checkpending];
  curimpl = [super new];
  return curimpl;
}

+ end
{
  if (curimpl) {

    id prev = curimpl;

    /* set global back to nil to end implemenation context */
    curimpl = nil;

    return prev;
  } else {
    fatalError("Class end outside class definition context");
    return nil;
  }
}

+ currentName
{
  return [curimpl idName];
}

+ current
{
  return curimpl;
}


- addClsdisp:aMethod
{
  if (!clsdisptbl)
    clsdisptbl = [OrdCltn new];
  [clsdisptbl add:aMethod];
  return self;
}

- addNstdisp:aMethod
{
  if (!nstDispatchTable)
    nstDispatchTable = [OrdCltn new];
  [nstDispatchTable add:aMethod];
  return self;
}

- moddefs
{
  id sName;

  assert(isImplementation);

  if ([options onePfile]) {
    [Module uniqclass:self];
  } else {
    [Module addClsimpl:self];
  }

  sName = [interface superName];	/* can be nil */
  if (sName)
    [Module usesEntry:sName];

  /* add entries for this implementation to the Module */
  [Module definesEntry:name];

  return self;
}


/*
 * definition of the Objective C class struct
 * for compatibility with Stepstone's runtime
 * there are some fields that we don't use (attr & objcID)
 *
 * "isa" is struct _SHARED for Stepstone, but I like "id" better.
 *
 * stes apr29/97
 * Changed output of struct _PRIVATE to struct Class_PRIVATE.
 * You can then in a C source level debugger cast to the 
 * struct _PRIVATE of some class.
 */

- (STR) privname
{
  return [[String sprintf:"%s_PRIVATE", [name str]] str];
}

- (STR) otbname
{
  return [[String sprintf:"%s_OTB", [name str]] str];
}

- privOn:stream
{
  [stream echo:"struct %s {", [self privname]];
/* instance variables */
  [LineTag forceTag];
  [interface ivarnodesOn:stream];
  [stream echo:"\n};\n"];

/* for debugging - this list should match the def above */
  if ([options comments])
    [interface ivarscommentOn:stream];

  if ([options otb]) {
    [stream echo:"struct %s {\n", [self otbname]];
    [stream echo:"\tstruct %s *ptr;\n", [self privname]];
    [Module otbvars:stream];
    [stream echo:"};\n"];
  }
  return self;
}

/*
 * stes sep/1/97
 * definition of "struct _SHARED" for a class
 * moved this from module.m to classdef.m because it is now possibly
 * different for each class (because of support for class variables)
 * 
 */

/*
 * definition of the Objective C class struct
 * for compatibility with Stepstone's runtime, our
 * extensions come after the "extensions" comment
 *
 * Note that the reference to struct _SLT is a pointer,
 * so we can do this without defining the type.
 * (objcc defines it in a header, and does NOT generate it).
 * So for cross-compiles, we have to do the same.
 */

- (STR) sharedname
{
  return [[String sprintf:"%s_SHARED", [name str]] str];
}

#if 0
static char *sharedType = "\n"
"struct _SHARED {\n"
/* keep this compatible with objcc */
"\tid isa;\n"
"\tid clsSuper;\n"
"\tchar 		*clsName;\n"
"\tchar		*clsTypes;\n"
"\tshort 		clsSizInstance;\n"
"\tshort 		clsSizDict;\n"
"\tstruct _SLT 	*clsDispTable;\n"
"\tlong 		clsStatus;\n"
"\tstruct modDescriptor *clsMod;\n"
/* not all objcc 4.x have clsVersion */
"\tunsigned	clsVersion;\n"
/* extensions */
"\tid clsCats;\n"
"};\n"
;

#endif

- shartypeOn:stream
{
  STR thisShared = [interface sharedname];

  [stream echo:"struct %s {\n", thisShared];

  /* keep this compatible with objcc's struct _SHARED */

  [stream echo:"\tid isa;\n"];
  [stream echo:"\tid clsSuper;\n"];
  [stream echo:"\tchar *clsName;\n"];
  [stream echo:"\tchar *clsTypes;\n"];
  [stream echo:"\tshort clsSizInstance;\n"];
  [stream echo:"\tshort clsSizDict;\n"];
  [stream echo:"\tstruct _SLT *clsDispTable;\n"];
  [stream echo:"\tlong clsStatus;\n"];
  [stream echo:"\tstruct modDescriptor *clsMod;\n"];

  /* not all objcc 4.x have clsVersion */

  [stream echo:"\tunsigned clsVersion;\n"];

  /* Portable Object Copmiler extensions begin here */

  [stream echo:"\tid clsCats;\n"];
  [stream echo:"\tid *clsGlbl;\n"];

  /* class variables */

  [self cvarnodesOn:stream];

  [stream echo:"\n};\n"];

  return self;
}

static void 
printslot(id stream, id aMethod, id className)
{
  STR s = [[aMethod fullname] str];
  STR t = [[aMethod funnameforclass:className] str];

  if ([options cplus]) {
    [stream echo:"\n{\"%s\",(id (*)(...))%s},", s, t];
  } else {
    [stream echo:"\n{\"%s\",(id (*)())%s},", s, t];
  }
}

/*
   * SGI cc doesn't like (aborts on) empty declarations
   * so always output at least {0,0}
 */

static void 
nullslot(id stream)
{
  if ([options cplus]) {
    [stream echo:"\n{(char*)0,(id (*)(...))0}"];
  } else {
    [stream echo:"\n{(char*)0,(id (*)())0}"];
  }
}


- (STR) _clsdisptblname
{
  if ([options onePfile]) {
    return "_clsDispatchTbl";
  } else {
    return [[String sprintf:"_%s_clsDispatchTbl", [name str]] str];
  }
}

- (STR) _nstDispatchTblName
{
  if ([options onePfile]) {
    return "_nstDispatchTbl";
  } else {
    return [[String sprintf:"_%s_nstDispatchTbl", [name str]] str];
  }
}

- clsdisptblsOn:stream
{
  int i, n;

  n = (clsdisptbl) ? [clsdisptbl size] : 0;

  [stream newline];
  [stream echo:"static struct _SLT %s[] ={", [self _clsdisptblname]];

  for (i = 0; i < n; i++)
    printslot(stream, [clsdisptbl at:i], [self idName]);
  nullslot(stream);
  [stream echo:"\n};\n"];
  [stream newline];

  return self;
}

- nstdisptblsOn:stream
{
  int i, n;

  n = (nstDispatchTable) ? [nstDispatchTable size] : 0;

  [stream newline];
  [stream echo:"static struct _SLT %s[] ={", [self _nstDispatchTblName]];
  for (i = 0; i < n; i++)
    printslot(stream, [nstDispatchTable at:i], [self idName]);
  nullslot(stream);
  [stream echo:"\n};\n"];
  [stream newline];

  return self;
}

- disptablesOn:stream
{
  STR sName = [name str];

  if ([options comments]) {
    [stream newline];
    [stream echo:"\n/*\n"];
    [stream echo:" * %s Dispatch tables \n", sName];
    [stream echo:" */\n"];
    [stream newline];
  }
  [self clsdisptblsOn:stream];
  [self nstdisptblsOn:stream];

  return self;
}

/* clsStatus : these must match objcrt.m */
#define CLS_FACTORY 0x1L
#define CLS_META 0x2L
#define CLS_INITIALIZED 0x4L
#define CLS_POSING 0x8L
#define CLS_MAPPED 0x10L
/* POC extensions begin here */
#define CLS_CATS 0x20L		/* obsolete, will go away */
#define CLS_CAT 0x40L

static void 
comment(id stream, STR comment)
{
  if ([options comments])
    [stream echo:comment];
  [stream newline];
}

- sharOn:stream
{
  long st;
  STR strsupername;
  STR strName = [name str];

  strsupername = ([interface superName]) ? [[interface superName] str] : NULL;

  if ([options comments]) {
    [stream newline];
    [stream echo:"\n/*\n"];
    [stream echo:" * struct for class %s \n", strName];
    [stream echo:" */\n"];
    [stream newline];
  }
/* name of struct */
  [stream echo:"struct %s ", [interface sharedname]];

  if (![options otb]) {
    [stream echo:" _%s = {\n", strName];
  } else {
    [stream echo:" _%s_OTB = {\n", strName];
  }

/* isa of our isa = our meta */
  [stream echo:"\t(id)&__%s,", strName];
  comment(stream, "/* isa */");

/* super of our isa = isa of our super */
/* in the -noShared case this field is set in BIND() */
  if (strsupername && [options sharedData]) {
    [stream echo:"\t(id)&_%s,", strsupername];
  } else {
    [stream echo:"\t0,"];	/* ends super hierarchy */
  }
  comment(stream, "/* clsSuper */");

/* clsName */
  [stream echo:"\t\"%s\",", strName];
  comment(stream, "/* clsName */");

/* clsTypes : no filer types yet */
  [stream echo:"\t0,"];
  comment(stream, "/* clsTypes */");

/* let C compiler compute size */
  [stream echo:"\tsizeof(struct %s_PRIVATE),", strName];
  comment(stream, "/* clsSizInstance */");

/* clsSizDict */
  [stream echo:"\t%d,", (nstDispatchTable) ? [nstDispatchTable size] : 0];
  comment(stream, "/* clsSizDict */");

/* _nstDispatchTbl */
  [stream echo:"\t%s,", [self _nstDispatchTblName]];
  comment(stream, "/* _nstDispatchTbl */");

/* clsStatus */
  st = CLS_FACTORY | CLS_POSING | CLS_CATS;
  if ([interface isCategory])
    st |= CLS_CAT;
  [stream echo:"\t%d,", st];
  comment(stream, "/* clsStatus */");

/* clsMod */
  [stream echo:"\t&%s,", [[Module moddescname] str]];
  comment(stream, "/* clsMod */");

/* clsVersion */
  [stream echo:"\t0,"];
  comment(stream, "/* clsVersion */");

/* clsCats - obsoleted */
  [stream echo:"\t(id)0,"];
  comment(stream, "/* clsCats */");

/* clsGlbl */
  [stream echo:"\t&%s,", strName];
  comment(stream, "/* clsGlbl */");

/* end of class struct */
  [stream echo:"};\n"];

  if ([options otb]) {
    STR z = strName;

    [stream newline];
    [stream echo:"struct OTB _%s={ (struct _PRIVATE*)&_%s_OTB,0,0 };\n", z, z];
  }
  /* output for -noShare and -share */
  [stream echo:"id %s %s(void) { return (id)&_%s; }\n",
   [[options bind] str], [self structfun], strName];
  if (![options otb]) {
    [stream echo:"id %s %s(void) { return _%s.clsSuper; }\n",
     [[options bind] str], [self superfun], strName];
  } else {
    [stream echo:"id %s %s(void) { return _%s_OTB.clsSuper; }\n",
     [[options bind] str], [self superfun], strName];
  }

  return self;
}

- metasharOn:stream
{
  long st;
  STR strsupername;
  STR strName = [name str];
  STR rootName = [[interface rootName] str];
  STR metaShared = [[interface rootInterface] sharedname];
  STR e = [[options bind] str];

  strsupername = ([interface superName]) ? [[interface superName] str] : NULL;

  assert(rootName);

  if ([options comments]) {
    [stream newline];
    [stream echo:"\n/*\n"];
    [stream echo:" * %s metaclass struct\n", strName];
    [stream echo:" */\n"];
    [stream newline];
  }
/* name of struct */
  [stream echo:"struct %s ", metaShared];

  if (![options otb]) {
    [stream echo:" __%s = {\n", strName];
  } else {
    [stream echo:" __%s_OTB = {\n", strName];
  }

  if ([options sharedData]) {
    /* isa of our meta = meta of Object */
    [stream echo:"\t(id)&__%s,", rootName];
  } else {
    [stream echo:"\t(id)0,"];	/* field set in BIND() */
  }
  comment(stream, "/* isa */");

  if ([options sharedData]) {
    /* super of our meta = meta of our super */
    if (strsupername) {
      [stream echo:"\t(id)&__%s,", strsupername];
    } else {
      /* stes jan25, 1997
       * The meta hierarchy is ended NOT by a '0'
       * but rather by making it point to the root class
       * itself which has a super of '0'.  This makes classes
       * respond to Object instance methods, such as |name|
       * (This was done wrong for objc1.1.1)
       * In other words: a search doesn't end with +name, but
       * with -name.
       */
      [stream echo:"\t(id)&_%s,", rootName];
    }
  } else {
    [stream echo:"\t(id)0,"];	/* field set in BIND() */
  }
  comment(stream, "/* clsSuper */");

/* clsName */
  [stream echo:"\t\"%s\",", strName];
  comment(stream, "/* clsName */");

/* clsTypes : no filer types yet */
  [stream echo:"\t0,"];
  comment(stream, "/* clsTypes */");

/* let C compiler compute size */
  [stream echo:"\tsizeof(struct %s),", [interface sharedname], strName];
  comment(stream, "/* clsSizInstance */");

/* clsSizDict */
  [stream echo:"\t%d,", (clsdisptbl) ? [clsdisptbl size] : 0];
  comment(stream, "/* clsSizDict */");

/* _nstDispatchTbl */
  [stream echo:"\t%s,", [self _clsdisptblname]];
  comment(stream, "/* _clsDispatchTbl */");

/* clsStatus */
  st = CLS_META | CLS_CATS;
  [stream echo:"\t%d,", st];
  comment(stream, "/* clsStatus */");

/* clsMod */
  [stream echo:"\t&%s,", [[Module moddescname] str]];
  comment(stream, "/* clsMod */");

/* clsVersion */
  [stream echo:"\t0,"];
  comment(stream, "/* clsVersion */");

/* clsCats - obsoleted */
  [stream echo:"\t(id)0,"];
  comment(stream, "/* clsCats */");

/* clsGlbl */
  [stream echo:"\t0,"];
  comment(stream, "/* clsGlbl */");

/* end of class struct */
  [stream echo:"};\n"];

  if ([options otb]) {
    STR z = strName;

    [stream newline];
    [stream echo:"struct OTB __%s={ (struct _PRIVATE*)&__%s_OTB,0,0 };\n", z, z];
  }
  /* output for -noShare and -share */
  [stream echo:"id %s %s(void) { return (id)&__%s; }\n",
   e, [self metastructfun], strName];
  if (![options otb]) {
    [stream echo:"id %s %s(void) { return __%s.clsSuper; }\n",
     e, [self metasuperfun], strName];
  } else {
    [stream echo:"id %s %s(void) { return __%s_OTB.clsSuper; }\n",
     e, [self metasuperfun], strName];
  }

  return self;
}

- globfuncallOn:stream
{
  STR t;
  STR s = [name str];
  id rootface = [interface rootInterface];
  id superface = [interface superInterface];

  /* this is called for each class in a module, inside OBJCBIND()
   * for the module.
   * allows us, in the -noShare case, to call a function and like
   * this fill out fields in the struct shared and meta of a class
   * (all necessary because problems with shared data with DLL's)
   */

  /* '0' ends super hierarchy, otherwise call function */
  if (superface) {
    t = [superface structfun];
    [stream echo:"\t_%s.clsSuper = %s();\n", s, t];
  }
  /* meta isa and clsSuper are never '0' */
  t = [rootface metastructfun];
  [stream echo:"\t__%s.isa = %s();\n", s, t];

  /* '0' does NOT end meta hierarchy (see comment above) */
  if (superface) {
    t = [superface metastructfun];
  } else {
    t = [rootface structfun];
  }
  [stream echo:"\t__%s.clsSuper = %s();\n", s, t];

  return self;
}

- prologueOn:stream
{
  STR s = [name str];
  STR t = ([interface superName]) ? [[interface superName] str] : NULL;
  STR r = [[interface rootName] str];
  STR superShared;
  STR thisShared = [interface sharedname];
  STR metaShared = [[interface rootInterface] sharedname];
  STR e = [[options bind] str];

  superShared = (t) ? [[interface superInterface] sharedname] : metaShared;

  if ([options comments]) {
    [stream newline];
    [stream echo:"\n/*\n"];
    [stream echo:" * %s implementation\n", s];
    [stream echo:" */\n"];
    [stream newline];
  }
/* forward declaration of class             */
/* already happened by the Class interface  */
/* the static global in the -noShared case, best no in interface */

  if ([options sharedData] == NO)
    [stream echo:"static id %s;\n", s];

/* forward declaration of meta of root class */
/* used for our struct _SHARED */
/* root class of root class is itself */

  [stream newline];
  if (![options otb]) {
    [stream echo:"extern struct %s __%s;\n", metaShared, r];
  } else {
    [stream echo:"extern struct OTB __%s;\n", r];
  }
  if ([options otb] || [options sharedData] == NO) {
    id rnf = [interface rootInterface];

    [stream echo:"extern id %s %s(void);\n", e, [rnf metastructfun]];
    [stream echo:"extern id %s %s(void);\n", e, [rnf metasuperfun]];
  }
/* forward declaration of class and metaclass structs     */
/* for superClass.  these are used for our struct _SHARED */
/* not needed if we're defining a rootClass (would be NULL) */
  if (t) {
    if (![options otb]) {
      [stream echo:"extern struct %s _%s;\n", superShared, t];
      [stream echo:"extern struct %s __%s;\n", metaShared, t];
    } else {
      [stream echo:"extern struct OTB _%s;\n", t];
      [stream echo:"extern struct OTB __%s;\n", t];
    }
    if ([options otb] || [options sharedData] == NO) {
      id snf = [interface superInterface];

      [stream echo:"extern id %s %s(void);\n", e, [snf structfun]];
      [stream echo:"extern id %s %s(void);\n", e, [snf metastructfun]];
      [stream echo:"extern id %s %s(void);\n", e, [snf superfun]];
      [stream echo:"extern id %s %s(void);\n", e, [snf metasuperfun]];
    }
  }
/* forward declaration of class and metaclass structs */
/* these are referred to in messages to "super"       */
  if (![options otb]) {
    [stream echo:"extern struct %s _%s;\n", thisShared, s];
    [stream echo:"extern struct %s __%s;\n", metaShared, s];
  } else {
    [stream echo:"extern struct OTB _%s;\n", s];
    [stream echo:"extern struct OTB __%s;\n", s];
  }
  if ([options otb] || [options sharedData] == NO) {
    [stream echo:"extern id %s %s(void);\n", e, [interface structfun]];
    [stream echo:"extern id %s %s(void);\n", e, [interface metastructfun]];
    [stream echo:"extern id %s %s(void);\n", e, [interface superfun]];
    [stream echo:"extern id %s %s(void);\n", e, [interface metasuperfun]];
  }
/* definition of "struct _PRIVATE" */
  [self privOn:stream];
  [stream newline];

  /* ended flag is needed in the context of implementations
   * for whether to emit a prologue or epilogue.
   * for interfaces it doesn't matter
   */

  prologueEmitted = YES;

  /* module defs for the references we have made in this method */
  [self moddefs];

  [LineTag forceTag];
  return self;
}

- superOn:stream
{
  if ([options otb] || [options sharedData] == NO) {
    STR x = [self superfun];

    [stream echo:"%s()", x];
  } else {
    STR x = [name str];

    [stream echo:"_%s.clsSuper", x];
  }
  return self;
}

- metasuperOn:stream
{
  if ([options otb] || [options sharedData] == NO) {
    STR x = [self metasuperfun];

    [stream echo:"%s()", x];
  } else {
    STR x = [name str];

    [stream echo:"__%s.clsSuper", x];
  }
  return self;
}

/******************************************************************************
 *
 *           A S C I I F I L E R   M E T H O D S
 *
 *****************************************************************************/

- superfileOutmsg
{
  id r = [Identifier str:"super"];
  id k = [Identifier str:"fileOutIdsFor"];
  id a = [Identifier str:"aFiler"];
  id xpr = [MessageExpression receiver:r keyw:k arg:a];

  return [[[Node new] add:xpr] add:[Token str:";\n"]];
}

- superfileInmsg
{
  id r = [Identifier str:"super"];
  id k = [Identifier str:"fileInIdsFrom"];
  id a = [Identifier str:"aFiler"];
  id xpr = [MessageExpression receiver:r keyw:k arg:a];

  return [[[Node new] add:xpr] add:[Token str:";\n"]];
}

- fileOutmsg:ivar
{
  id r = [Identifier str:"aFiler"];
  id k1 = [Identifier str:"fileOut"];
  id a1 = [[[Node new] add:[Token str:"&"]] add:ivar];
  id k2 = [Identifier str:"type"];
  id a2 = [Identifier str:"'@'"];
  id xpr = [MessageExpression receiver:r keyw:k1 arg:a1 keyw:k2 arg:a2];

  [xpr preventfwd:YES];
  return [[[Node new] add:xpr] add:[Token str:";\n"]];
}

- fileInmsg:ivar
{
  id r = [Identifier str:"aFiler"];
  id k1 = [Identifier str:"fileIn"];
  id a1 = [[[Node new] add:[Token str:"&"]] add:ivar];
  id k2 = [Identifier str:"type"];
  id a2 = [Identifier str:"'@'"];
  id xpr = [MessageExpression receiver:r keyw:k1 arg:a1 keyw:k2 arg:a2];

  [xpr preventfwd:YES];
  return [[[Node new] add:xpr] add:[Token str:";\n"]];
}

static id 
fileoutmeth(void)
{
  id method = [MethodInfo new];

  [method isFactory:NO];
  [method preventfwd:YES];
  [method resType:[Token str:"id"]];
  [method addComponent:[String str:"fileOutIdsFor:"]];
  [method addArgtype:[Token str:"id"]];
  [method addArgument:[Identifier str:"aFiler"]];
  /* add necessary info to class structures */
  [method implement];
  return method;
}

static id 
fileinmeth(void)
{
  id method = [MethodInfo new];

  [method isFactory:NO];
  [method preventfwd:YES];
  [method resType:[Token str:"id"]];
  [method addComponent:[String str:"fileInIdsFrom:"]];
  [method addArgtype:[Token str:"id"]];
  [method addArgument:[Identifier str:"aFiler"]];
  /* add necessary info to class structures */
  [method implement];
  return method;
}

- genfileOutcode:stream
{
  id m, stmt;
  id l = [Token str:"{"];
  id r = [Token str:"}"];
  id ivars, ivar, method = fileoutmeth();

  stmt = [Node new];

  /* generate call to super */
  if ([interface superName]) {
    [stmt add:[self superfileOutmsg]];
    ivars = [[interface allIdivars] eachElement];
    while ((ivar = [ivars next])) {
      /* make a copy without lineTag associated */
      [stmt add:[self fileOutmsg:[ivar copy]]];
    }
  }
  [stmt add:[Identifier str:"return (id)self;\n"]];
  stmt = [CompoundStatement lbrace:l body:stmt rbrace:r];

  m = [MethodImplementation new];
  [m add:[[MethodDeclarator new] method:method]];
  [m add:stmt];
  [m genOn:stream];

  return self;
}

- genfileIncode:stream
{
  id m, stmt;
  id l = [Token str:"{"];
  id r = [Token str:"}"];
  id ivar, ivars, method = fileinmeth();

  stmt = [Node new];

  if ([interface superName]) {
    [stmt add:[self superfileInmsg]];
    ivars = [[interface allIdivars] eachElement];
    while ((ivar = [ivars next])) {
      /* make a copy without lineTag associated */
      [stmt add:[self fileInmsg:[ivar copy]]];
    }
  }
  [stmt add:[Identifier str:"return (id)self;\n"]];
  stmt = [CompoundStatement lbrace:l body:stmt rbrace:r];

  m = [MethodImplementation new];
  [m add:[[MethodDeclarator new] method:method]];
  [m add:stmt];
  [m genOn:stream];

  return self;
}

- filercodeOn:stream
{
  [stream newline];
  [stream newline];

  if ([options comments]) {
    [stream echo:"\n/*\n"];
    [stream echo:" *           A S C I I F I L E R   M E T H O D S\n"];
    [stream echo:" */\n"];
    [stream newline];
  }
  if ([[interface allIdivars] isEmpty] == NO) {
    [self genfileOutcode:stream];
    [stream newline];
    [self genfileIncode:stream];
  }
  return self;
}

/******************************************************************************
 *
 * epilogueOn
 *
 *****************************************************************************/

- globalOn:stream
{
  STR s = [name str];

  [stream newline];

  /* assign struct SHARED to class global */
  /* make this static in -noShared case.  */

  if ([options sharedData] == NO)
    [stream echo:"static "];
  [stream echo:"id %s = (id)&_%s;\n", s, s];

  /* in both -noShared and -shared case, output a globalfun() */
  /* (in theory only needed for -noShare) but can do it always */

  [stream echo:"id %s *%s(void) { return & %s; }\n",
   [[options bind] str], [self globalfun], s];

  return self;
}

- epilogueOn:stream
{
  STR s = [name str];

  /* this must be done in the curimpl context */
  curimpl = self;
  if ([options filerCode])
    [self filercodeOn:stream];
  curimpl = nil;

  [self disptablesOn:stream];
  [self sharOn:stream];
  [self metasharOn:stream];
  [self globalOn:stream];

  if ([options comments]) {
    [stream newline];
    [stream echo:"\n/*\n"];
    [stream echo:" * End of %s implementation\n", s];
    [stream echo:" */\n"];
    [stream newline];
  }
  [LineTag forceTag];
  return self;
}

- makecurivars
{
  [interface makeCurrent];
  return self;
}

- clearcurivars
{
  [interface clearCurrent];
  return self;
}

- startimpl:stream
{
  /* case when we thought we were an interface but at once we find
   * a method implementation (as opposed to a prototype)
   * So promote ourselves to an implementation (and interface at
   * the same time) and translate the prologue before the method impl.
   */

  if (isImplementation == NO) {

    /* this is to avoid mistakes with class implementations
     * without interface declaration part (which we allow,
     * but do not allow if the super class isn't specified)
     */

    if (interface == self && superName == nil) {
      fatalError("root class without interface not allowed");
    }
    /* another mistake is to put a method impl in an interface */
    if ([self isintftoken]) {
      fatalError("method implementation inside @interface");
    }
    [self isImplementation:YES];
    [self genOn:stream];
  }
  return self;
}

- genOn:stream
{
  if (isImplementation) {

    if (!interfaceEmitted)
      [self interfaceOn:stream];

    if (prologueEmitted) {
      [self epilogueOn:stream];
    } else {
      [self prologueOn:stream];
    }
  } else {
    if (!interfaceEmitted)
      [self interfaceOn:stream];
  }

  return self;
}

@end

