
/*
 * Portable Object Compiler (c) 1997.  All Rights Reserved.
 * $Id: linetag.m,v 1.18 1998/10/06 15:23:44 stes Exp $
 */

/*
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include "ocstring.h"
#include "identif.h"
#include "linetag.h"
#include "options.h"
#include "output.h"
#include <assert.h>
#include <string.h>		/* strlen */

@implementation LineTag

/*****************************************************************************
 *
 * Amateurish parsing code for line tags.
 * Alternative is to use a mode in lex code, but I decided against that,
 * because it should be feasible to "clean up" the scan.l code by hand-coding
 * here some C for the (simple) case of line directives.
 *
 * Utility function to check whether "aDirective" is something like
 *
 * 	# lineNo "filename" ... \n
 * or
 * 	# line lineNo "filename" ... \n
 *
 * If yes, returns by reference the lineNumber and a string object
 * (the filename)
 *
 * NOTE: since HPUX cc is also emitting # lineNo (without fileName)
 * have to support this.
 *
 ****************************************************************************/

static STR 
getlineno(STR s, int *lineNo)
{
  id tmp;
  STR p = s;
  int c, n = 0;

  while ((c = *p++) && '0' <= c && c <= '9')
    n++;

  tmp = [String chars:s count:n];
  *lineNo = [tmp asInt];
  tmp = [tmp free];

  return p - 1;
}

static BOOL 
getfilename(STR s, id *fileName)
{
  STR p = s;
  int c, n = 0;

  while ((c = *p++) && c != '"')
    n++;

  /* no \" in filenames (yet) (contrary to scan.l:STRING_TOKEN) */
  if (c == '"') {
    *fileName = [String chars:s count:n];
    return YES;
  } else {
    return NO;
  }
}

static BOOL 
isLine(STR aDirective, int *lineNo, id *fileName)
{
  int c;
  STR p = aDirective;

  /* match the hash sign */
  c = *p++;
  if (c == '#') {
    c = *p++;
  } else {
    return NO;
  }

  /* optional whitespace */
  while (c && (c == ' ' || c == '\t'))
    c = *p++;

  /* accept both "# no name" and "# line no name" */
  if (c == 'l') {
    STR key = "ine";		/* l already consumed */

    if (strncmp(p, key, strlen(key)) == 0) {
      p += strlen(key);
      c = *p++;
    } else {
      return NO;
    }
  }
  /* optional whitespace */
  while (c && (c == ' ' || c == '\t'))
    c = *p++;

  /* get the line number */
  if (c && '0' <= c && c <= '9') {
    p = getlineno(p - 1, lineNo);
    c = *p++;
  } else {
    return NO;
  }

  /* optional whitespace */
  while (c && (c == ' ' || c == '\t'))
    c = *p++;

  /* get the filename */
  if (c && c == '"') {
    return getfilename(p, fileName);
    /* ignore everything else on the line after the fName */
  } else {
    /* # lineNo style */
    *fileName = nil;
    /* ignore everything else on the line after the fName */
    return YES;
  }
}

/* pragmas are ignored and simply placed again on the output 
 * however, a few do have side-effects.
 * for example, "tcc" is declaring builtin-types via pragmas
 * and so does the Stepstone compiler.
 */

static void 
pragtoktype(void)
{
  char *t;
  char *sep = " \t\n\r";

  /* #pragma token TYPE foo # */
  /* #pragma OCbuiltInType */
  /* this can also be done with -builtintype on the command line */

  t = strtok(NULL, sep);
  if (!t)
    return;
  [[[Identifier str:t] markdecl:YES] deftypename];
}

static void 
pragtokvariety(void)
{
  char *t;
  char *sep = " \t\n\r";

  /* #pragma token VARIETY foo # */
  /* this can also be done with -builtintype on the command line */

  t = strtok(NULL, sep);
  if (!t)
    return;
  if (strcmp(t, "signed") == 0 || strcmp(t, "unsigned") == 0) {
    t = strtok(NULL, sep);
    if (!t)
      return;
  }
  [[[Identifier str:t] markdecl:YES] deftypename];
}

static void 
pragtokproc(void)
{
  char *t = NULL;
  char *sep = " \t\n\r";

  /* #pragma token PROC (blahblah) foo # */
  /* this can also be done with -builtinfunction on the command line */

  while (1) {
    id prv = [String str:t];

    t = strtok(NULL, sep);
    if (!t)
      return;
    if (strcmp(t, "#") == 0) {
      [[Identifier str:[prv str]] defbuiltinfun];
      return;
    }
    if (prv)
      [prv free];
  }
}

static void 
pragtokproc2(void)
{
  char *t = NULL;
  char *sep = " \t\n\r";

  /* #pragma OCbuiltInFctn foo */

  t = strtok(NULL, sep);
  if (!t)
    return;
  [[Identifier str:t] defbuiltinfun];
}

static STR 
dopragmas(STR s)
{
  char *t;
  char *sep = " \t";
  STR buf = [[String str:s] str];	/* make a copy */

  t = strtok(buf, sep);
  if (!t)
    return s;

  if (strcmp(t, "#pragma") == 0) {
    t = strtok(NULL, sep);
    if (!t)
      return s;
    if (strcmp(t, "token") == 0) {
      t = strtok(NULL, sep);
      if (!t)
	return s;
      if (strcmp(t, "TYPE") == 0) {
	pragtoktype();
	return s;
      }
      if (strcmp(t, "VARIETY") == 0) {
	pragtokvariety();
	return s;
      }
      if (strcmp(t, "EXP") == 0)
	return s;
      if (strcmp(t, "FUNC") == 0)
	return s;
      if (strcmp(t, "PROC") == 0) {
	pragtokproc();
	return s;
      }
    }
    if (strcmp(t, "printLine") == 0)
      return t + strlen("printLine") + 1;
    if (strcmp(t, "OCbuiltInFctn") == 0) {
      pragtokproc2();
      return "";
    }
    if (strcmp(t, "OCbuiltInType") == 0) {
      pragtoktype();
      return "";
    }
  }
  /* unrecognized pragma */
  return s;
}


/*****************************************************************************
 *
 * Creation
 *
 ****************************************************************************/


static id currentTag;

- next
{
  return next;
}

- setNext:aTag
{
  next = aTag;
  return self;
}

- fileName:(STR)fName lineNo:(int)no directive:s
{
  fileName = fName;
  lineNo = no;
  directive = s;

  /* maintain linked list to be able to traverse ranges of tags */
  /* can't use collection (grows over 32K for large files) */
  if (currentTag)
    [currentTag setNext:self];

  /* depend on the list nil terminated */
  currentTag = [self setNext:nil];

  return self;
}

+ new
{
  int no;
  STR fName;

  /* we get a +new before yylex() is called */
  if (currentTag) {
    fName = [currentTag fileName];
    no = [currentTag lineNo] + 1;
  } else {
    id iFile = [options inFile];

    fName = (iFile) ? [iFile str] : "stdin";
    no = 1;
  }

  return [[super new] fileName:fName lineNo:no directive:nil];
}

+ str:(STR)aDirective
{
  int no;
  id s, d;
  STR fName;
  id current = currentTag;

  if (isLine(aDirective, &no, &s)) {
    if (s) {
      fName = [s strCopy];
      s = [s free];
    } else {
      /* "# lineNo" style */
      fName = [current fileName];
      if (!fName)
	warnError("# lineNo without filename");
    }

    d = nil;
    return [[super new] fileName:fName lineNo:no directive:d];
  } else {

    /* cc pragma's, indents etc. */

    if (aDirective)
      aDirective = dopragmas(aDirective);

    fName = [current fileName];
    no = [current lineNo] + 1;
    d = [String str:aDirective];
    return [[super new] fileName:fName lineNo:no directive:d];
  }
}

+ currentTag
{
  return currentTag;
}

- (int) lineNo
{
  return lineNo;
}

- (STR) fileName
{
  return fileName;
}

+ (int) curlineno
{
  return [currentTag lineNo];
}

+ (STR) curfilename
{
  return [currentTag fileName];
}

static id outCurrent;
static BOOL forceOutput;

static void 
gentagOn(int no, STR fName, id stream)
{
  id format = [options tagformat];

  [stream echo:[format str], no, fName];
  [stream newline];
}

+ outputOn:stream
{
  STR fName;

  if ([options srcfilename]) {
    fName = [[options srcfilename] str];
  } else {
    fName = "stdin";
  }

  if ([options lineTags])
    gentagOn(1, fName, stream);
  outCurrent = [[super new] fileName:fName lineNo:1 directive:nil];

  return self;
}

+ forceTag
{
  forceOutput = YES;
  return self;
}

- directiveOn:stream
{
  /* the pragma includes a newline */
  if (directive) {
    /* [self outputOn:stream]; */
    [stream echoId:directive];
  }
  return self;
}

- outputOn:stream
{
  if ([options lineTags])
    gentagOn(lineNo, fileName, stream);
  return self;
}

+ (BOOL) areSuccessors:a:b
{
  if ([a lineNo] + 1 == [b lineNo]) {
    return strcmp([a fileName], [b fileName]) == 0;
  } else {
    return NO;
  }
}

- (BOOL) comesBefore:b
{
  id nextTag = next;

  while (nextTag) {
    if (nextTag == b)
      return YES;
    nextTag = [nextTag next];
  }

  return NO;
}

+ outputOn:stream until:offset
{
  /* note that we cannot assert that offset > outCurrent */
  /* indeed, it's possible to dump a piece of code (like */
  /* in a struct for @defs) with linetags that are lower */
  /* than outCurrent (ivars for the @defs being declared */
  /* in some header file, long before the @defs usage)   */

  /* just output whitespace if there was no newline */
  if (forceOutput == NO && outCurrent == offset) {
    [stream echo:" "];
    return self;
  }
  [stream newline];

  /* a single newline is sufficient here if they're successors */
  if (forceOutput || [self areSuccessors:outCurrent:offset] == NO) {

    if ([outCurrent comesBefore:offset]) {
      id aTag = outCurrent;

      /* only do pragma's etc. - no consecutive newlines */
      while (aTag != offset) {
	aTag = [aTag next];
	if (aTag)
	  [aTag directiveOn:stream];
      }
    }
    /* just one lineTag */
    [offset outputOn:stream];
  }
  forceOutput = NO;
  outCurrent = offset;

  return self;
}

+ finishOn:stream
{
  id aTag = outCurrent;

  [stream newline];

  do {
    aTag = [aTag next];
    if (aTag)
      [aTag directiveOn:stream];
  } while (aTag);

  [stream newline];
  return self;
}

@end

