
/*
 * Portable Object Compiler (c) 1997.  All Rights Reserved.
 * $Id: methinfo.m,v 1.23 1998/10/06 15:23:45 stes Exp $
 */

/*
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "config.h"
#include "methinfo.h"
#include "classdef.h"
#include "ocstring.h"
#include "set.h"
#include "dictnary.h"
#include "token.h"
#include "identif.h"
#include "ordcltn.h"
#include "output.h"
#include "options.h"
#include "linetag.h"
#include "compstat.h"
#include "msgxpr.h"
#include <assert.h>

@implementation MethodInfo

static id methodPrototypes;

static void 
warnnotfound(id aName)
{
  if ([options warnundefined]) {
    char b[8192];

    sprintf(b, "argument types for method %s default to 'id'", [aName str]);
    warnError(b);
  }
}

static void 
warntypeconflict(id aName)
{
  if ([options warntypeconflict]) {
    char b[8192];

    sprintf(b, "Conflicting type casts for selector '%s'", [aName str]);
    warnError(b);
  }
}

- free
{
  return nil;
}

- prototype
{
  id p, aName;

  if (!methodPrototypes)
    methodPrototypes = [Dictionary new];

  aName = [self fullname];

  if ((p = [methodPrototypes atKey:aName])) {
    if ([self typeconflicts:p])
      warntypeconflict(aName);
  } else {
    [methodPrototypes atKey:aName put:self];
  }

  return self;
}

- implement
{
  id aClass = [ClassDefinition current];

  /* implementing implicitely prototypes */
  /* also - prototype will check types   */

  [self prototype];

  /* add to the class implementation */

  if (aClass) {
    if ([self isFactory]) {
      [aClass addClsdisp:self];
    } else {
      [aClass addNstdisp:self];
    }
  } else {
    fatalError("method implementation outside class definition");
  }

  return self;
}

+ findprotofor:aName
{
  return [methodPrototypes atKey:aName];
}

- defaultargs
{
  [self resType:nil];
  /* arguments have been defined as "id" too */
  assert([arguments size] == [argumentTypes size]);
  return self;
}

+ findproto:aMethod
{
  id m;
  id aName = [aMethod fullname];

  if (!(m = [self findprotofor:aName])) {
    warnnotfound(aName);
    [aMethod defaultargs];
    [aMethod prototype];
    m = [self findprotofor:aName];
    assert(m != nil);
  }
  return m;
}


- (BOOL) isEqual:aMethod
{
  /* +new is for us same selector as -new */
  return [name isEqual:[aMethod fullname]];
}

- (BOOL) typeconflicts:b
{
  return [self istypeEqual:b] == NO;
}

- (BOOL) istypeEqual:b
{
  if (![[self resType] isEqual:[b resType]])
    return NO;
  if (argumentTypes == nil)
    return [b argumentTypes] == nil;
  if ([b argumentTypes] == nil)
    return argumentTypes == nil;
  if (![argumentTypes isEqual:[b argumentTypes]])
    return NO;
  return YES;
}

- (unsigned) typehash
{
  unsigned h;
  unsigned cnt = [argumentTypes size];

  h = cnt & ([[self resType] hash] << 8);
  if (cnt == 1)
    h &= ([[argumentTypes at:0] hash] << 16);
  if (cnt == 2)
    h &= ([[argumentTypes at:1] hash] << 24);
  return h;
}


- fullname
{
  return name;
}

- (STR) uscorename
{
  /* this depends on selector name only (not class) */

  if (uscorename) {
    return [uscorename str];
  } else {
    char *q;

    /* convert colons to underscores */
    uscorename = [name copy];
    for (q = [uscorename str]; *q != 0; q++) {
      if (*q == ':')
	*q = '_';
    }
    return [uscorename str];
  }
}

- preventfwd:(BOOL)flag
{
  preventfwd = flag;
  return self;
}

- (BOOL) canforward
{
  if (preventfwd)
    return NO;
  return ellipsis == nil;	/* no support for var.args. yet */
}

- (BOOL) calcselptr
{
  if (preventfwd == NO && [options selptr]) {
    int n = [argumentTypes size];

    if (![[self resType] isPtr])
      return NO;
    if (n > 4)
      return NO;
    if (n > 3 && ![[argumentTypes at:3] isPtr])
      return NO;
    if (n > 2 && ![[argumentTypes at:2] isPtr])
      return NO;
    if (n > 1 && ![[argumentTypes at:1] isPtr])
      return NO;
    if (n > 0 && ![[argumentTypes at:0] isPtr])
      return NO;
    return YES;
  } else {
    return NO;
  }
}

- (BOOL) isselptr
{
  if (didselptr) {
    return isselptr;
  } else {
    didselptr = YES;
    return isselptr = [self calcselptr];
  }
}

- (STR) fwdname
{
  if (fwdname) {
    return [fwdname str];	/* can do this, since depends on selector only */
  } else {
    fwdname = [String sprintf:"%s_fwd", [self uscorename]];
    return [fwdname str];
  }
}

- (STR) dispname
{
  if (dispname) {
    return [dispname str];	/* can do this, since depends on selector only */
  } else {
    dispname = [String sprintf:"%s_disp", [self uscorename]];
    return [dispname str];
  }
}

- (STR) argstructname
{
  if (argstructname) {
    return [argstructname str];
  } else {
    argstructname = [String sprintf:"%s_args", [self uscorename]];
    return [argstructname str];
  }
}

- funnameforclass:aClassName
{
  /* 'c_' prefix denotes class/factory method
   * 'i_' stands for instance method
   * Include classname so that we can have multiple methods
   * implementations
   * (for different classes) in same file.
   */

  if ([self isFactory]) {
    return [String sprintf:"c_%s_%s", [aClassName str], [self uscorename]];
  } else {
    return [String sprintf:"i_%s_%s", [aClassName str], [self uscorename]];
  }
}

- isFactory:(BOOL)flag
{
  isFactory = flag;
  return self;
}

- (BOOL) isFactory
{
  return isFactory;
}

- resType
{
  return (resultType) ? resultType : (resultType = [Token str:"id"]);
}

- (BOOL) isresultvoid
{
  return [[self resType] isVoid];
}

- resType:aNode
{
  resultType = aNode;
  return self;
}


- addComponent:aString
{
  if (!name)
    name = [String new];
  [name concatSTR:[aString str]];
  return self;
}

- addColon
{
  assert(name);
  [name concatSTR:":"];
  return self;
}

- addEllipsis:aToken
{
  ellipsis = aToken;
  return self;
}

- addArgtype:aNode
{
  if (!argumentTypes)
    argumentTypes = [OrdCltn new];
  [argumentTypes add:aNode];
  /* protect ourselves against missing |processCurrent| calls */
  assert([argumentTypes size] < 100);
  return self;
}

- argumentTypes
{
  return argumentTypes;
}

- addId
{
  return [self addArgtype:[Token newToken:"id"]];
}

- addArgument:aNode
{
  if (!arguments)
    arguments = [OrdCltn new];
  [arguments add:aNode];
  /* protect ourselves against missing |processCurrent| calls */
  assert([arguments size] < 100);
  return self;
}

static void 
warnwith(STR name, int expected, int found)
{
  char b[8192];

  sprintf(b, "method %s%d followed by %d arguments", name, expected, found);
  warnError(b);
}

- checkwith
{
  id firstEl;
  int expected, n;
  static id likeWith;

  if (!likeWith) {
    likeWith = [Set add:[String str:"with:"]];
  }
  /* stes 1/98
   * suggestion Marc Boschma marcb@c3.telstra-mm.net.au
   * to count number of arguments for +with: method
   * we could also automatically generate the count
   * so that the user doesn't have to tell us the number
   * of arguments, but that would be non-portable
   * across Objective-C compilers
   */

  if (![likeWith contains:[self fullname]])
    return self;

  /* way we deal with comma-list arguments is kind-o-fishy since
   * it counts here as one colon argument +with:oneArg where
   * oneArg is actually in itself a Node (expression) with the
   * commalist.  we want the firstElement of the list
   */

  assert([arguments size] == 1);
  firstEl = [[arguments lastElement] subnodes];
  n = [firstEl size];
  assert(n >= 1);
  n = (n - 1) / 2;		/* commas are in there ... */

  expected = [[firstEl firstElement] asInt];	/* gives 0 if symbolic */

  /* it might be symbolic (expected=0), then we cant check */
  if (expected && (n != expected)) {
    warnwith([[self fullname] str], expected, n);
  }
  return self;
}


static id stdmsgargs;

- stdmsgargs
{
  if (stdmsgargs)
    return stdmsgargs;
  stdmsgargs = [OrdCltn new];
  [stdmsgargs add:[Identifier str:"self"]];
  [stdmsgargs add:[Identifier str:"_cmd"]];
  return stdmsgargs;
}

- stdmsgtypes
{
  id aClass, s;
  id stdmsgtypes = [OrdCltn new];	/* types are different each class */

  /* 11/97 - changed id type to the current class type */
  aClass = [ClassDefinition current];
  if (![options otb]) {
    s = [String sprintf:"struct %s *", [aClass privname]];
  } else {
    s = [String sprintf:"struct %s *", [aClass otbname]];
  }
  [stdmsgtypes add:[Identifier string:s]];
  [stdmsgtypes add:[Identifier str:"SEL"]];
  return stdmsgtypes;
}

- grepLocals:aCompound
{
  [aCompound addLocals:[self stdmsgargs] types:[self stdmsgtypes]];
  [aCompound addInitvars:[self stdmsgargs]];

  if (arguments) {
    [aCompound addLocals:arguments types:argumentTypes];
    [aCompound addInitvars:arguments];
  }
  return self;
}


- resulttypeOn:stream
{
  [[self resType] taglessOn:stream];
  return self;
}

- castOn:stream
{
  [self resulttypeOn:stream];

  /* function pointer */
  [stream echo:" (*) "];

  /* argument types */
  [stream echo:"("];
  [self typesOn:stream];
  [stream echo:")"];

  return self;
}

- typesOn:stream
{
  int i, n;

  n = (argumentTypes) ? [argumentTypes size] : 0;

  /* first the "self" and "_cmd" types */
  [stream echo:"id,SEL"];

  for (i = 0; i < n; i++) {
    [stream echo:","];
    [[argumentTypes at:i] taglessOn:stream];
  }

  if (ellipsis) {
    [stream echo:","];
    [ellipsis taglessOn:stream];
  }
  return self;
}

- idselfcmdOn:stream
{
  [stream echo:"id self,SEL _cmd"];
  return self;
}

- selfcmdOn:stream
{
  STR t;
  id aClass;

  /* first the "self" and "_cmd" types 

   * stes - feb 3, 1997
   * because we changed for Microsoft C the def of "id"
   * as emitted by objc (see Module.m)
   * we need to introduce a cast here.
   * This isn't bad actually, since, when we will do more than 1
   * object per file, this would be the place to the cast.
   *
   * stes - apr29, 1997
   * Prepended name of class to _PRIVATE.
   *
   * 11/97 our philosophy is that self just an 'id' variable,
   * but many people ask for instance variable access in a debugger
   * so we now type this here as a struct PRIVATE
   * (as opposed to how it was before, where the cast was done
   * only at the time of variable translation)
   */

  aClass = [ClassDefinition current];
  t = ([options otb]) ? [aClass otbname] : [aClass privname];
  [stream echo:"struct %s *self,SEL _cmd", t];

  return self;
}

- protosOn:stream
{
  int i, m, n;

  m = (arguments) ? [arguments size] : 0;
  n = (argumentTypes) ? [argumentTypes size] : 0;

  if (m != n) {
    warnError("internal error: number arguments != number types");
  }
  for (i = 0; i < n; i++) {
    [stream echo:","];
    [[argumentTypes at:i] taglessOn:stream];
    [[arguments at:i] taglessOn:stream];
  }

  if (ellipsis) {
    [stream echo:","];
    [ellipsis taglessOn:stream];
  }
  return self;
}

- genargsOn:stream
{
  int i, n;

  n = (arguments) ? [arguments size] : 0;

  /* always preceded by self and SEL so output a comma */

  for (i = 0; i < n; i++) {
    [stream echo:","];
    [[arguments at:i] genOn:stream];
  }

  return self;
}

- arguments:a
{
  arguments = a;
  return self;
}

- dispatchCopy
{
  int i, n;
  id newargs = [OrdCltn new];

  n = (arguments) ? [arguments size] : 0;
  for (i = 0; i < n; i++) {
    id arg;
    STR s = [[arguments at:i] str];

    if (s) {
      arg = [String sprintf:"args->%s", s];
      [newargs add:[Identifier string:arg]];
    }
  }

  /* copy where we replace arguments by args->x as needed in dispatch func */
  return [[self copy] arguments:newargs];
}

- genargsintostructOn:stream
{
  int i, n;

  n = (arguments) ? [arguments size] : 0;
  for (i = 0; i < n; i++) {
    [stream echo:"args."];
    [[arguments at:i] taglessOn:stream];
    [stream echo:"="];
    [[arguments at:i] taglessOn:stream];
    [stream echo:";"];
  }
  if (n)
    [stream newline];
  return self;
}

- funprotoOn:stream
{
  id s;

  [stream echo:"static "];
  [self resulttypeOn:stream];
  s = [self funnameforclass:[ClassDefinition currentName]];
  [stream echo:" %s(", [s str]];
  [self selfcmdOn:stream];
  [self protosOn:stream];
  [stream echo:")"];

  return self;
}

- fwdstubOn:stream
{
#if 0
  [self argstructOn:stream];
  [stream newline];
#endif

  [stream echo:"static "];
  [self resulttypeOn:stream];
  [stream echo:" %s(", [self fwdname]];
  [self idselfcmdOn:stream];
  [self protosOn:stream];
  [stream echo:")\n"];
  [stream echo:"{\n"];
  [stream echo:"struct %s args;\n", [self argstructname]];
  [self genargsintostructOn:stream];
  [stream echo:"fwdmsg(self,_cmd,&args,(ARGIMP)%s);\n", [self dispname]];
  if (![self isresultvoid])
    [stream echo:"return args._retval;\n"];
  [stream echo:"}\n"];
  [stream newline];
  return self;
}


- argstructOn:stream
{
  int i;

  [stream echo:"struct %s {\n", [self argstructname]];
  if (![self isresultvoid]) {
    [self resulttypeOn:stream];
    [stream echo:" _retval;\n"];
  } else {
    [stream echo:"int _retval;\n"];	/* assures non-empty struct */
  }
  for (i = 0; i < [arguments size]; i++) {
    [[argumentTypes at:i] taglessOn:stream];
    [stream echo:" "];
    [[arguments at:i] taglessOn:stream];
    [stream echo:";\n"];
  }
  /* do nothing for ellipsis */
  [stream echo:"};\n"];
  return self;
}

- dispfunOn:stream
{
  id x;

  assert([self canforward]);

  [stream echo:"static void %s(", [self dispname]];
  [self idselfcmdOn:stream];
  [stream echo:",struct %s *args)\n", [self argstructname]];
  [stream echo:"{\n"];
  if (![self isresultvoid])
    [stream echo:"args->_retval = "];
  x = [Identifier str:"self"];	/* rcvr */
  x = [[MessageExpression new] dispatch:x method:[self dispatchCopy]];
  [x genmsgOn:stream inSuper:NO];
  [stream echo:";\n"];
  [stream echo:"}\n"];
  return self;
}

- grepScope:aMethod
{
  /* don't process all subnodes, esp. not the keyw. selector itself
   * as it might have the naem of an instance variable */
  if (arguments)
    [arguments elementsPerform:_cmd with:aMethod];
  return self;
}

@end

